import dataclasses
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Iterable, List

from oarepo_upload_cli.config import Config

from ..types import JsonType


class RecordSource(ABC):
    """
    Describes a source that is used to generate records.
    """

    def __init__(self, config: Config) -> None:
        self._config = config

    @abstractmethod
    def get_records(
        self, modified_after: datetime = None, modified_before: datetime = None
    ) -> Iterable["SourceRecord"]:
        """
        Provides a generator that returns records within given timestamps.
        If no timestamps are given, returns all records. The timestamps are not timezone
        aware and are in UTC.
        """

    @abstractmethod
    def get_records_count(
        self, modified_after: datetime = None, modified_before: datetime = None
    ) -> int:
        """
        Approximates the size of a collection of records being returned.
        The timestamps are not timezone aware and are in UTC.
        """


@dataclasses.dataclass
class SourceRecord:
    """
    Describes a source record.
    """

    record_id: str
    datetime_modified: datetime
    deleted: bool
    metadata: JsonType
    files: List["SourceRecordFile"]


@dataclasses.dataclass
class SourceRecordFile(ABC):
    """
    Represent a source record file.
    """

    key: str
    content_type: str
    datetime_modified: datetime
    metadata: JsonType

    @abstractmethod
    def get_reader(self):
        pass



import requests

from oarepo_upload_cli.exceptions import (
    ExceptionMessage,
    RepositoryCommunicationException,
)


class InvenioConnection:
    def __init__(self, auth):
        self._auth = auth
        self._json_headers = {"Content-Type": "application/json"}

    def get(self, url, *, params=None):
        return self.send_request("get", url=url, params=params)

    def post(self, url, *, json=None):
        return self.send_request("post", url=url, json=json)

    def put(self, url, *, json=None, headers=None, data=None):
        return self.send_request("put", url=url, json=json, headers=headers, data=data)

    def delete(self, url):
        return self.send_request("delete", url=url)

    def send_request(self, http_verb, **kwargs):
        try:
            request_method = getattr(globals()["requests"], http_verb)
            headers = kwargs.pop("headers", self._json_headers)
            res = request_method(
                verify=False, auth=self._auth, headers=headers, **kwargs
            )
            res.raise_for_status()
        except requests.ConnectionError as conn_err:
            raise RepositoryCommunicationException(
                ExceptionMessage.ConnectionError, conn_err
            ) from conn_err
        except requests.HTTPError as http_err:
            raise RepositoryCommunicationException(
                ExceptionMessage.HTTPError, http_err, res.text, url=kwargs["url"]
            ) from http_err
        except Exception as err:
            raise RepositoryCommunicationException(str(err), err) from err

        return res



from requests.auth import AuthBase


class TokenAuthentication(AuthBase):
    """
    Implements a custom token-based authentication.
    """

    def __init__(self, header_field: str, token: str):
        self.header_field = header_field
        self.token = token

    def __call__(self, req):
        req.headers[self.header_field] = f"{self.token}"
        return req



from abc import abstractmethod
from json import JSONDecodeError
from typing import Any, Dict, Optional, Union

import requests

from oarepo_upload_cli.base.repository import RepositoryClient, RepositoryRecord
from oarepo_upload_cli.base.source import SourceRecord
from oarepo_upload_cli.exceptions import (
    ExceptionMessage,
    RepositoryCommunicationException,
)
from oarepo_upload_cli.invenio.connection import InvenioConnection
from oarepo_upload_cli.invenio.record import InvenioRepositoryRecord


class InvenioRepositoryClient(RepositoryClient):
    record_class = InvenioRepositoryRecord

    def __init__(self, config):
        super().__init__(config)
        self.connection = InvenioConnection(config.auth)

    @abstractmethod
    def get_id_query(self, source_record_id: str) -> Dict[str, str]:
        raise NotImplementedError("ID query not implemented")

    @abstractmethod
    def get_last_modification_date(self) -> Optional[str]:
        raise NotImplementedError("Last modification date not implemented")

    def get_record(self, record: SourceRecord) -> RepositoryRecord:
        params = self.get_id_query(record.record_id)

        res = self.connection.get(url=self._config.collection_url, params=params)

        res_payload = res.json()
        hits = res_payload["hits"]["hits"]

        if hits:
            return self.record_class(
                self.connection,
                self._config.collection_url,
                hits[0],
                self._config.record_modified_field_name,
                self._config.file_modified_field_name,
            )

    def create_record(self, record: SourceRecord) -> RepositoryRecord:
        res = self.connection.post(
            url=self._config.collection_url, json=record.metadata
        )
        return self.record_class(
            self.connection,
            self._config.collection_url,
            res.json(),
            self._config.record_modified_field_name,
            self._config.file_modified_field_name,
        )

    def delete_record(self, record: RepositoryRecord):
        record: InvenioRepositoryRecord  # can delete only my record
        self.connection.delete(url=record.self_url)

    def _get_aggregation(self, *path: str) -> Union[Any, None]:
        """
        Sends a request to the given repository URL. Tries to acquire the data from the response determined by the given path.

        Returns the data or prints an error with the description what happened.
        """

        try:
            url = self._config.collection_url
            res = self.connection.get(url)
        except requests.ConnectionError as conn_err:
            raise RepositoryCommunicationException(
                ExceptionMessage.ConnectionError, conn_err
            ) from conn_err
        except requests.exceptions.HTTPError as http_err:
            raise RepositoryCommunicationException(
                ExceptionMessage.HTTPError, http_err, res.text, url=url
            ) from http_err
        except Exception as err:
            raise RepositoryCommunicationException(str(err), err) from err

        try:
            content = res.json()
        except JSONDecodeError as serialization_err:
            raise RepositoryCommunicationException(
                ExceptionMessage.JSONContentNotSerializable, serialization_err
            ) from serialization_err

        content_element = content
        for path_element in ["aggregations"] + list(path):
            if path_element not in content_element:
                raise KeyError(
                    f"Path element {path_element} from {path} not found in data {content}"
                )
            content_element = content_element[path_element]
        return content_element



from datetime import datetime
from typing import Callable, Optional

from oarepo_upload_cli.base.repository import RepositoryClient, RepositoryRecord
from oarepo_upload_cli.base.source import RecordSource, SourceRecord
from oarepo_upload_cli.config import Config


def noop(*args, **kwargs):
    pass


class Uploader:
    def __init__(
        self, config: Config, source: RecordSource, repository: RepositoryClient
    ):
        self.config = config
        self.source = source
        self.repository = repository

    def upload(
        self,
        modified_after: datetime = None,
        modified_before: datetime = None,
        callback: Callable[[SourceRecord, int, int, str], None] = noop,
    ):
        """
        :param modified_after:      datetime when to start uploads
        :param modified_before:     datetime when to stop uploads
        :param callback: function(source_record: SourceRecord, current_record_count, approximate_records_count, message)
        :return: number of uploaded/deleted/modified records
        """

        if not modified_before:
            modified_before = datetime.utcnow()
        if not modified_after:
            modified_after = self.repository.get_last_modification_date()

        approximate_records_count = self.source.get_records_count(
            modified_after, modified_before
        )

        for record_cnt, source_record in enumerate(
            self.source.get_records(modified_after, modified_before)
        ):

            def local_callback(msg):
                callback(
                    source_record,
                    record_cnt,
                    approximate_records_count,
                    msg,
                )

            repository_record = self._create_update_record_metadata(
                source_record,
                local_callback,
            )

            if not repository_record:
                continue

            self._create_update_record_files(
                source_record, repository_record, local_callback
            )

    def _create_update_record_metadata(
        self, source_record, callback
    ) -> Optional[RepositoryRecord]:
        repository_record = self.repository.get_record(source_record)

        if source_record.deleted:
            if repository_record:
                self.repository.delete_record(repository_record)
                callback("deleted")
            return None

        if not repository_record:
            repository_record = self.repository.create_record(source_record)
            callback("created")

        elif repository_record.datetime_modified < source_record.datetime_modified:
            repository_record.update_metadata(source_record.metadata)
            callback("updated")

        return repository_record

    @staticmethod
    def _create_update_record_files(source_record, repository_record, callback):
        processed_keys = set()
        for f in source_record.files:
            if repository_record.create_update_file(f):
                callback(f"{f.key} uploaded")
            processed_keys.add(f.key)

        for f in list(repository_record.files.values()):
            if f.key not in processed_keys:
                repository_record.delete_file(f)
                callback(f"{f.key} deleted")



import configparser
import os

from oarepo_upload_cli.auth.bearer_auth import BearerAuthentication


class Config:
    def __init__(self, init_file_path=None, *overrides):
        config = configparser.ConfigParser()
        if init_file_path and os.path.exists(init_file_path):
            config.read(init_file_path)

        for section in ("authentication", "repository", "entrypoints"):
            if not config.has_section(section):
                config.add_section(section)

        self.config = config
        for override in overrides:
            self.override(*override)

    @property
    def auth(self):
        return BearerAuthentication(self.bearer_token)

    @property
    def bearer_token(self):
        return self.ensure_defined(
            "authentication", "token", "REPOSITORY_UPLOADER_BEARER_TOKEN"
        )

    @property
    def collection_url(self):
        return self.ensure_defined(
            "repository", "collection_url", "REPOSITORY_UPLOADER_COLLECTION_URL"
        )

    @property
    def file_modified_field_name(self):
        return (
            self.config["repository"].get("file_modified_field")
            or os.getenv("REPOSITORY_UPLOADER_FILE_MODIFIED_FIELD_NAME")
            or "dateModified"
        )

    @property
    def record_modified_field_name(self):
        return (
            self.config["repository"].get("record_modified_field")
            or os.getenv("REPOSITORY_UPLOADER_RECORD_MODIFIED_FIELD_NAME")
            or "dateModified"
        )

    @property
    def source_name(self):
        return self.ensure_defined(
            "entrypoints", "source", "REPOSITORY_UPLOADER_SOURCE"
        )

    @property
    def repository_name(self):
        return self.ensure_defined(
            "entrypoints", "repository", "REPOSITORY_UPLOADER_REPOSITORY"
        )

    def override(self, section, option, value):
        if value:
            self.config[section][option] = value

    def ensure_defined(self, section, config_key, os_environ_key):
        ret = self.config[section].get(config_key) or os.getenv(os_environ_key)
        if not ret:
            raise Exception(
                f"Please supply {config_key} to init file or set {os_environ_key} environment variable"
            )
        return ret



import json
from typing import Dict, Optional

from oarepo_upload_cli.base.repository import RepositoryClient, RepositoryRecord
from oarepo_upload_cli.base.source import SourceRecord, SourceRecordFile
from oarepo_upload_cli.types import JsonType


class DryRepositoryRecord(RepositoryRecord):
    def __init__(self, record: SourceRecord):
        self.record = record

    @property
    def datetime_modified(self):
        return self.record.datetime_modified

    @property
    def record_id(self):
        return self.record.record_id

    @property
    def files(self):
        return []

    def create_update_file(self, file: SourceRecordFile) -> bool:
        print(f"Creating or updating file {file.key} of record {self.record_id}")
        return True

    def create_file(self, file: SourceRecordFile):
        print(f"Creating file {file.key} of record {self.record_id}")

    def update_file(self, file: SourceRecordFile):
        print(f"Updating file {file.key} of record {self.record_id}")

    def delete_file(self, file: SourceRecordFile):
        print(f"Deleting file {file.key} of record {self.record_id}")

    def update_metadata(self, new_metadata: Dict[str, JsonType]):
        print(f"Updating metadata of record {self.record_id}: {new_metadata}")


class DryRepositoryClient(RepositoryClient):
    def get_id_query(self, source_record_id: str) -> Dict[str, str]:
        return {}

    def get_last_modification_date(self) -> Optional[str]:
        return None

    def get_record(self, source_record: SourceRecord) -> RepositoryRecord:
        return None

    def create_record(self, source_record: SourceRecord) -> RepositoryRecord:
        print(
            f"\nCreating {source_record.id} {json.dumps(source_record.metadata, ensure_ascii=False, indent=4)}"
        )
        return DryRepositoryRecord(source_record)

    def delete_record(self, record: RepositoryRecord):
        print(f"\nDeleting {record.record_id}")



import importlib_metadata

from .exceptions import ExceptionMessage


class EntryPointsLoader:
    def __init__(self):
        self._group = "oarepo_upload_cli.dependencies"

    def load_entry_point(self, name: str, default=None):
        """
        Tries to load an entry point given by argument value or
        """
        eps = importlib_metadata.entry_points(group=self._group)
        for ep in eps:
            if ep.name == name:
                return ep.load()
        if default:
            return default
        raise ValueError(
            f"{ExceptionMessage.EntryPointNotProvided.value}: group '{self._group}', name '{name}'"
        )



from enum import Enum


class ExceptionMessage(str, Enum):
    """
    Collection of string representing exception error messages.
    """

    ConnectionError = "Network problem has occurred"
    EntryPointNotProvided = "Entry point not provided"
    HTTPError = "HTTP error has occured"
    JSONContentNotSerializable = "Response could not be serialized"
    MultipleEntryPoints = "Multiple entry points present, can not choose one"


class EntryPointNotFoundException(Exception):
    """
    Raise when an entry point is not defined.
    """

    def __init__(self):
        super().__init__(ExceptionMessage.EntryPointNotProvided.value)


class RepositoryCommunicationException(Exception):
    """
    Raised when there is a problem with requesting data from repository happens.
    """

    def __init__(self, message, error, payload=None, url=None):
        super().__init__(message)
        self._error = error
        self._payload = payload

        if url:
            print(f"Exception on a request with url: {url}.")

        if payload:
            print(f"Exception's payload: {payload}.")



import dataclasses
from datetime import datetime
from functools import cached_property
from typing import Dict
from urllib.parse import urljoin

from oarepo_upload_cli.base.repository import RepositoryFile, RepositoryRecord
from oarepo_upload_cli.base.source import SourceRecordFile
from oarepo_upload_cli.invenio.connection import InvenioConnection
from oarepo_upload_cli.types import JsonType


@dataclasses.dataclass
class InvenioRepositoryFile(RepositoryFile):
    metadata: Dict[str, JsonType]
    file_modified_field_name: str

    @property
    def datetime_modified(self):
        if self.file_modified_field_name in (self.metadata.get("metadata") or {}):
            return datetime.fromisoformat(
                self.metadata["metadata"][self.file_modified_field_name]
            )


@dataclasses.dataclass
class InvenioRepositoryRecord(RepositoryRecord):
    connection: InvenioConnection
    base_url: str
    metadata: Dict[str, JsonType]
    record_modified_field_name: str
    file_modified_field_name: str

    @cached_property
    def files(self) -> Dict[str, InvenioRepositoryFile]:
        return {
            x["key"]: InvenioRepositoryFile(x["key"], x, self.file_modified_field_name)
            for x in self.connection.get(self.link_url("files")).json()["entries"]
        }

    @property
    def record_id(self):
        return self.metadata["id"]

    def link_url(self, key):
        base_url = self.base_url
        if not base_url.endswith("/"):
            base_url += "/"
        return urljoin(base_url, self.metadata["links"][key])

    @property
    def self_url(self):
        return self.link_url("self")

    @property
    def files_url(self):
        return self.link_url("files")

    @property
    def datetime_modified(self):
        return datetime.fromisoformat(
            self.metadata["metadata"][self.record_modified_field_name]
        )

    def update_metadata(self, new_metadata: Dict[str, JsonType]):
        self.metadata = self.connection.put(url=self.self_url, json=new_metadata).json()

    def create_update_file(self, file: SourceRecordFile) -> bool:
        existing_file: RepositoryFile
        if file.key in self.files:
            existing_file = self.files[file.key]
            if (
                existing_file.datetime_modified
                and existing_file.datetime_modified >= file.datetime_modified
            ):
                # no need to update
                return False

            # invenio can not perform update, so at first delete and then create
            self.delete_file(file)

        self.create_file(file)
        return True

    def create_file(self, file: SourceRecordFile):
        # raises exception on error
        self.connection.post(url=self.files_url, json=[{"key": file.key}])
        self.update_file(file)

    def update_file(self, file: SourceRecordFile):
        url = f"{self.files_url}/{file.key}"
        content_url = f"{self.files_url}/{file.key}/content"
        commit_url = f"{self.files_url}/{file.key}/commit"

        # put metadata
        self.connection.put(url=url, json=file.metadata)

        # upload data
        self.connection.put(
            url=content_url,
            headers={"Content-Type": file.content_type},
            data=file.get_reader(),
        )

        # commit
        self.connection.post(commit_url)

        # reread the metadata to make sure they have been uploaded
        repository_file = InvenioRepositoryFile(
            key=file.key,
            metadata=self.connection.get(url),
            file_modified_field_name=self.file_modified_field_name,
        )
        self.files[file.key] = repository_file

    def delete_file(self, file: SourceRecordFile):
        url = f"{self.files_url}/{file.key}"
        self.connection.delete(url)
        self.files.pop(file.key, None)



from pathlib import Path

import arrow
import click
from tqdm import tqdm

from oarepo_upload_cli.base.source import SourceRecord
from oarepo_upload_cli.config import Config
from oarepo_upload_cli.dry_run import DryRepositoryClient
from oarepo_upload_cli.entry_points_loader import EntryPointsLoader
from oarepo_upload_cli.uploader import Uploader


@click.command()
@click.option(
    "--config",
    "config_name",
    help="Configuration entry point path of config implementation.",
)
@click.option(
    "--source",
    "source_name",
    help="Configuration entry point path of source implementation.",
)
@click.option(
    "--repository",
    "repository_name",
    help="Configuration entry point path of repository client implementation.",
)
@click.option("--token", "token", help="Bearer token")
@click.option("--collection-url", "collection_url", help="Collection url")
@click.option(
    "--record-modified-field",
    "record_modified_field",
    help="Path to field containing record modification date",
)
@click.option(
    "--file-modified-field",
    "file_modified_field",
    help="Path to field containing file modification date",
)
@click.option(
    "--modified_after",
    help="Timestamp that represents date after modification. "
    "If not specified, the last updated timestamp from repository will be used.",
)
@click.option(
    "--modified_before", help="Timestamp that represents date before modification."
)
@click.option("--dry/--no-dry", help="Dry run - do not upload anything")
@click.option("--suppress-warnings/--no-suppress-warnings")
def main(
    config_name,
    source_name,
    repository_name,
    modified_after,
    modified_before,
    dry,
    suppress_warnings,
    token,
    collection_url,
    record_modified_field,
    file_modified_field,
) -> None:
    if suppress_warnings:
        import warnings

        warnings.filterwarnings("ignore")

    # -----------------
    # - Configuration -
    # -----------------
    ep_loader = EntryPointsLoader()
    config_class = ep_loader.load_entry_point(
        config_name,
        default=Config,
    )
    config = config_class(str(Path.home() / ".repository-uploader.ini"))
    config.override("authentication", "token", token)
    config.override("entrypoints", "source", source_name)
    config.override("entrypoints", "repository", repository_name)
    config.override("repository", "collection_url", collection_url)
    config.override("repository", "record_modified_field", record_modified_field)
    config.override("repository", "file_modified_field", file_modified_field)

    # --------------------------------
    # - Source and repository client -
    # --------------------------------
    source = ep_loader.load_entry_point(config.source_name)(config)
    if dry:
        repository = DryRepositoryClient(config)
    else:
        repository = ep_loader.load_entry_point(config.repository_name)(config)

    # --------------
    # - Timestamps -
    # --------------
    if modified_before:
        modified_before = arrow.get(modified_before).datetime.replace(tzinfo=None)

    if modified_after:
        modified_after = arrow.get(modified_after).datetime.replace(tzinfo=None)

    # ----------
    # - Upload -
    # ----------

    progress_bar = tqdm()

    def callback(source_record: SourceRecord, cnt, approximate_count, message):
        if progress_bar.total != approximate_count:
            progress_bar.total = approximate_count
        progress_bar.set_description(
            f"{source_record.id} {source_record.datetime_modified} {message}"
        )
        progress_bar.update(cnt)

    uploader = Uploader(config, source, repository)
    uploader.upload(modified_after, modified_before, callback)


if __name__ == "__main__":
    main()



import dataclasses
from abc import ABC, abstractmethod
from enum import Enum
from typing import Dict, Optional

from oarepo_upload_cli.base.source import SourceRecord, SourceRecordFile
from oarepo_upload_cli.config import Config
from oarepo_upload_cli.types import JsonType


@dataclasses.dataclass
class RepositoryFile:
    key: str


class RepositoryRecord(ABC):
    @property
    @abstractmethod
    def datetime_modified(self):
        pass

    @property
    @abstractmethod
    def record_id(self):
        pass

    @property
    @abstractmethod
    def files(self) -> Dict[str, RepositoryFile]:
        return {}

    @abstractmethod
    def create_update_file(self, file: SourceRecordFile) -> bool:
        pass

    @abstractmethod
    def create_file(self, file: SourceRecordFile):
        pass

    @abstractmethod
    def update_file(self, file: SourceRecordFile):
        pass

    @abstractmethod
    def delete_file(self, file: SourceRecordFile):
        """
        Tries to delete a given file of a given record by its key.
        """

    @abstractmethod
    def update_metadata(self, new_metadata: Dict[str, JsonType]):
        """
        Perform actualization of a given records metadata.
        """


class RepositoryClient(ABC):
    def __init__(self, config: Config):
        self._config = config

    @abstractmethod
    def get_id_query(self, source_record_id: str) -> Dict[str, str]:
        pass

    @abstractmethod
    def get_last_modification_date(self) -> Optional[str]:
        pass

    def get_record(self, source_record: SourceRecord) -> RepositoryRecord:
        """
        Creates a record in the repository with the given metadata.

        Returns created record metadata.
        """

    def create_record(self, source_record: SourceRecord) -> RepositoryRecord:
        """
        Creates a record in the repository with the given metadata.

        Returns created record metadata.
        """

    def delete_record(self, record: RepositoryRecord):
        """
        Tries to delete a given record.
        """


class FileStatus(str, Enum):
    """
    Based on: https://github.com/inveniosoftware/invenio-records-resources/blob/5335294dade21decea0f527022d96e12e1ffad52/invenio_records_resources/services/files/schema.py#L115
    """

    COMPLETED = "completed"
    PENDING = "pending"
